@@ -31,6 +31,7 @@
 #include <ell/ell.h>
 
 #include "ell/tls-private.h"
+#include "ell/asn1-private.h"
 #include "ell/cert-private.h"
 
 static void test_tls10_prf(const void *data)
@@ -214,7 +215,7 @@ static struct l_cert *load_cert_file(const char *filename)
 	char *label;
 	struct l_cert *cert;
 
-	der = l_pem_load_file(filename, 0, &label, &len);
+	der = l_pem_load_file(filename, &label, &len);
 	if (!der)
 		return NULL;
 
@@ -228,11 +229,12 @@ static void test_certificates(const void *data)
 {
 	struct l_queue *cacert;
 	struct l_queue *wrongca;
+	struct l_queue *wrongca2;
 	struct l_queue *twocas;
+	struct l_cert *expired;
+	struct l_queue *mixedcas;
 	struct l_certchain *chain;
-	struct l_certchain *chain2;
-	struct l_certchain *chain3;
-	struct l_certchain *chain4;
+	struct l_certchain *expiredchain;
 
 	cacert = l_pem_load_certificate_list(CERTDIR "cert-ca.pem");
 	assert(cacert && !l_queue_isempty(cacert));
@@ -240,60 +242,123 @@ static void test_certificates(const void *data)
 	wrongca = l_pem_load_certificate_list(CERTDIR "cert-intca.pem");
 	assert(wrongca && !l_queue_isempty(wrongca));
 
+	wrongca2 = l_pem_load_certificate_list(CERTDIR "cert-server.pem");
+	assert(wrongca2 && !l_queue_isempty(wrongca2));
+
 	twocas = l_pem_load_certificate_list(CERTDIR "cert-chain.pem");
 	assert(twocas);
 
+	mixedcas = l_pem_load_certificate_list(CERTDIR "cert-chain.pem");
+	assert(mixedcas);
+	expired = load_cert_file(CERTDIR "cert-expired.pem");
+	assert(expired);
+	l_queue_push_tail(mixedcas, expired);
+
 	chain = l_pem_load_certificate_chain(CERTDIR "cert-server.pem");
 	assert(chain);
 
+	expiredchain = l_pem_load_certificate_chain(CERTDIR "cert-expired.pem");
+	assert(expiredchain);
+
 	assert(!l_certchain_verify(chain, wrongca, NULL));
 	assert(l_certchain_verify(chain, cacert, NULL));
 	assert(l_certchain_verify(chain, NULL, NULL));
 	assert(l_certchain_verify(chain, twocas, NULL));
 
-	chain2 = l_pem_load_certificate_chain(CERTDIR "cert-chain.pem");
-	assert(chain2);
+	l_certchain_free(chain);
 
-	assert(!l_certchain_verify(chain2, wrongca, NULL));
-	assert(l_certchain_verify(chain2, cacert, NULL));
-	assert(l_certchain_verify(chain2, NULL, NULL));
-	assert(l_certchain_verify(chain2, twocas, NULL));
+	chain = l_pem_load_certificate_chain(CERTDIR "cert-chain.pem");
+	assert(chain);
 
-	chain3 = certchain_new_from_leaf(
-			load_cert_file(CERTDIR "cert-server.pem"));
-	certchain_link_issuer(chain3,
-			load_cert_file(CERTDIR "cert-entity-int.pem"));
-	certchain_link_issuer(chain3,
-			load_cert_file(CERTDIR "cert-intca.pem"));
-	certchain_link_issuer(chain3,
-			load_cert_file(CERTDIR "cert-ca.pem"));
-	assert(chain3);
-
-	assert(!l_certchain_verify(chain3, wrongca, NULL));
-	assert(!l_certchain_verify(chain3, cacert, NULL));
-	assert(!l_certchain_verify(chain3, NULL, NULL));
-	assert(!l_certchain_verify(chain3, twocas, NULL));
-
-	chain4 = certchain_new_from_leaf(
-			load_cert_file(CERTDIR "cert-entity-int.pem"));
-	certchain_link_issuer(chain4,
-			load_cert_file(CERTDIR "cert-intca.pem"));
-	certchain_link_issuer(chain4,
-			load_cert_file(CERTDIR "cert-ca.pem"));
-	assert(chain4);
-
-	assert(!l_certchain_verify(chain4, wrongca, NULL));
-	assert(l_certchain_verify(chain4, cacert, NULL));
-	assert(l_certchain_verify(chain4, NULL, NULL));
-	assert(l_certchain_verify(chain4, twocas, NULL));
+	assert(!l_certchain_verify(chain, wrongca2, NULL));
+	assert(l_certchain_verify(chain, cacert, NULL));
+	assert(l_certchain_verify(chain, NULL, NULL));
+	assert(l_certchain_verify(chain, twocas, NULL));
 
 	l_certchain_free(chain);
-	l_certchain_free(chain2);
-	l_certchain_free(chain3);
-	l_certchain_free(chain4);
+
+	chain = certchain_new_from_leaf(
+			load_cert_file(CERTDIR "cert-server.pem"));
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-entity-int.pem"));
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-intca.pem"));
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-ca.pem"));
+	assert(chain);
+
+	assert(!l_certchain_verify(chain, wrongca, NULL));
+	assert(!l_certchain_verify(chain, cacert, NULL));
+	assert(!l_certchain_verify(chain, NULL, NULL));
+	assert(!l_certchain_verify(chain, twocas, NULL));
+
+	l_certchain_free(chain);
+
+	chain = certchain_new_from_leaf(
+			load_cert_file(CERTDIR "cert-entity-int.pem"));
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-intca.pem"));
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-ca.pem"));
+	assert(chain);
+
+	assert(!l_certchain_verify(chain, wrongca2, NULL));
+	assert(l_certchain_verify(chain, cacert, NULL));
+	assert(l_certchain_verify(chain, NULL, NULL));
+	assert(l_certchain_verify(chain, twocas, NULL));
+	assert(l_certchain_verify(chain, mixedcas, NULL));
+
+	l_certchain_free(chain);
+	l_queue_destroy(cacert, (l_queue_destroy_func_t) l_cert_free);
+
+	cacert = l_pem_load_certificate_list(CERTDIR "cert-ca2.pem");
+	assert(cacert && !l_queue_isempty(cacert));
+
+	chain = certchain_new_from_leaf(
+			load_cert_file(CERTDIR "cert-no-keyid.pem"));
+	assert(chain);
+
+	assert(!l_certchain_verify(chain, wrongca, NULL));
+	assert(l_certchain_verify(chain, cacert, NULL));
+	assert(l_certchain_verify(chain, NULL, NULL));
+	assert(!l_certchain_verify(chain, twocas, NULL));
+
+	certchain_link_issuer(chain,
+			load_cert_file(CERTDIR "cert-ca2.pem"));
+
+	assert(!l_certchain_verify(chain, wrongca, NULL));
+	assert(l_certchain_verify(chain, cacert, NULL));
+	assert(l_certchain_verify(chain, NULL, NULL));
+	assert(!l_certchain_verify(chain, twocas, NULL));
+	assert(!l_certchain_verify(chain, mixedcas, NULL));
+
+	assert(!l_certchain_verify(expiredchain, NULL, NULL));
+
+	l_certchain_free(chain);
+	l_certchain_free(expiredchain);
 	l_queue_destroy(cacert, (l_queue_destroy_func_t) l_cert_free);
 	l_queue_destroy(wrongca, (l_queue_destroy_func_t) l_cert_free);
+	l_queue_destroy(wrongca2, (l_queue_destroy_func_t) l_cert_free);
 	l_queue_destroy(twocas, (l_queue_destroy_func_t) l_cert_free);
+	l_queue_destroy(mixedcas, (l_queue_destroy_func_t) l_cert_free);
+}
+
+static void test_ec_certificates(const void *data)
+{
+	struct l_queue *cacert;
+	struct l_certchain *chain;
+
+	cacert = l_pem_load_certificate_list(CERTDIR "ec-cert-ca.pem");
+	assert(cacert && !l_queue_isempty(cacert));
+
+	chain = l_pem_load_certificate_chain(CERTDIR "ec-cert-server.pem");
+	assert(chain);
+
+	assert(l_certchain_verify(chain, cacert, NULL));
+	assert(l_certchain_verify(chain, NULL, NULL));
+
+	l_certchain_free(chain);
+	l_queue_destroy(cacert, (l_queue_destroy_func_t) l_cert_free);
 }
 
 struct tls_conn_test {
@@ -502,7 +567,6 @@ static void tls_debug_cb(const char *str, void *user_data)
 static void test_tls_with_ver(const struct tls_conn_test *test,
 				uint16_t server_ver, uint16_t client_ver)
 {
-	bool auth_ok;
 	struct tls_test_state s[2] = {
 		{
 			.ready = false,
@@ -556,16 +620,57 @@ static void test_tls_with_ver(const struct tls_conn_test *test,
 	if (getenv("TLS_DEBUG"))
 		l_tls_set_debug(s[1].tls, tls_debug_cb, "client", NULL);
 
-	auth_ok = l_tls_set_auth_data(s[0].tls, test->server_cert_path,
-					test->server_key_path,
-					test->server_key_passphrase);
-	assert(auth_ok);
-	auth_ok = l_tls_set_auth_data(s[1].tls, test->client_cert_path,
-					test->client_key_path,
-					test->client_key_passphrase);
-	assert(auth_ok);
-	assert(l_tls_set_cacert(s[0].tls, test->server_ca_cert_path));
-	assert(l_tls_set_cacert(s[1].tls, test->client_ca_cert_path));
+	if (test->server_cert_path || test->server_key_path) {
+		struct l_certchain *server_cert =
+			l_pem_load_certificate_chain(test->server_cert_path);
+		struct l_key *server_key =
+			l_pem_load_private_key(test->server_key_path,
+						test->server_key_passphrase,
+						NULL);
+		if (!server_key) {
+			l_info("* Some kernel versions do not automatically\n"
+				"* load the pkcs8_key_parser module. If the\n"
+				"* system running test has not loaded this\n"
+				"* module, a failure here is likely. Running\n"
+				"* \"modprobe pkcs8_key_parser\" may correct\n"
+				"* this issue.\n");
+			exit(1);
+		}
+
+		assert(server_key);
+		assert(server_cert);
+
+		assert(l_tls_set_auth_data(s[0].tls, server_cert, server_key));
+	}
+
+	if (test->client_key_path || test->client_cert_path) {
+		struct l_certchain *client_cert =
+			l_pem_load_certificate_chain(test->client_cert_path);
+		struct l_key *client_key =
+			l_pem_load_private_key(test->client_key_path,
+						test->client_key_passphrase,
+						NULL);
+		assert(client_cert);
+		assert(client_key);
+
+		assert(l_tls_set_auth_data(s[1].tls, client_cert, client_key));
+	}
+
+	if (test->server_ca_cert_path) {
+		struct l_queue *server_ca =
+			l_pem_load_certificate_list(test->server_ca_cert_path);
+
+		assert(server_ca);
+		assert(l_tls_set_cacert(s[0].tls, server_ca));
+	}
+
+	if (test->client_ca_cert_path) {
+		struct l_queue *client_ca =
+			l_pem_load_certificate_list(test->client_ca_cert_path);
+
+		assert(client_ca);
+		assert(l_tls_set_cacert(s[1].tls, client_ca));
+	}
 
 	if (test->client_domain_mask)
 		l_tls_set_domain_mask(s[1].tls, test->client_domain_mask);
@@ -836,26 +941,11 @@ static void test_tls_suite_test(const void *data)
 	test_tls_with_ver(&test, 0, 0);
 }
 
-static int read_int_from_file(const char *path)
-{
-	int ret;
-	FILE *file;
-
-	file = fopen(path, "r");
-	if (!file)
-		return 0;
-
-	if (fscanf(file, "%i", &ret) < 1)
-		ret = 0;
-
-	fclose(file);
-	return ret;
-}
-
 int main(int argc, char *argv[])
 {
 	unsigned int i;
-	int maxkeys;
+	uint32_t maxkeys;
+	int r;
 
 	l_test_init(&argc, &argv);
 
@@ -879,8 +969,10 @@ int main(int argc, char *argv[])
 	l_test_add("TLS 1.2 PRF with SHA512", test_tls12_prf,
 			&tls12_prf_sha512_0);
 
-	if (l_key_is_supported(L_KEY_FEATURE_RESTRICT))
+	if (l_key_is_supported(L_KEY_FEATURE_RESTRICT)) {
 		l_test_add("Certificate chains", test_certificates, NULL);
+		l_test_add("ECDSA Certificates", test_ec_certificates, NULL);
+	}
 
 	if (!l_getrandom_is_supported()) {
 		printf("getrandom missing, skipping TLS connection tests...\n");
@@ -888,8 +980,7 @@ int main(int argc, char *argv[])
 	}
 
 	if (!l_cipher_is_supported(L_CIPHER_DES3_EDE_CBC) ||
-			!l_cipher_is_supported(L_CIPHER_AES_CBC) ||
-			!l_cipher_is_supported(L_CIPHER_ARC4)) {
+			!l_cipher_is_supported(L_CIPHER_AES_CBC)) {
 		printf("Needed ciphers missing, "
 				"skipping TLS connection tests...\n");
 		goto done;
@@ -902,10 +993,9 @@ int main(int argc, char *argv[])
 		goto done;
 	}
 
-	maxkeys = read_int_from_file(getuid() > 0 ?
-					"/proc/sys/kernel/keys/maxkeys" :
-					"/proc/sys/kernel/keys/root_maxkeys");
-	if (maxkeys && maxkeys < 2000)
+	r = l_sysctl_get_u32(&maxkeys, "/proc/sys/kernel/keys/%s",
+				getuid() > 0 ? "maxkeys" : "root_maxkeys");
+	if (!r && maxkeys < 2000)
 		printf("Running sysctl kernel.keys.%s=2000 is recommended\n",
 			getuid() > 0 ? "maxkeys" : "root_maxkeys");
 
@@ -962,6 +1052,9 @@ int main(int argc, char *argv[])
 		struct tls_bulk_encryption_algorithm *alg = suite->encryption;
 		bool supported;
 
+		if (l_str_has_prefix(suite->name, "TLS_ECDHE_ECDSA"))
+			continue;
+
 		if (alg->cipher_type == TLS_CIPHER_AEAD)
 			supported = l_aead_cipher_is_supported(alg->l_aead_id);
 		else
